@@ -4,21 +4,12 @@
  *
  * Copyright (C) 2007  Kristian Hoegsberg <krh@bitplanet.net>
  *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software Foundation,
- * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+ * This library is licensed under the GNU Lesser General Public License (LGPL),
+ * version 2.1 or later. See the file COPYING.LIB in the distribution for
+ * details.
  */
 
+#include <ctype.h>
 #include <stdlib.h>
 #include <string.h>
 #include <stdio.h>
@@ -117,6 +108,13 @@ default_bus_reset_handler(raw1394handle_t handle, unsigned int gen)
 	return 0;
 }
 
+static int
+is_fw_device_name(char *name)
+{
+	return strncmp(name, FW_DEVICE_PREFIX, strlen(FW_DEVICE_PREFIX)) == 0
+	       && isdigit(name[strlen(FW_DEVICE_PREFIX)]);
+}
+
 static int
 scan_devices(fw_handle_t handle)
 {
@@ -142,8 +140,7 @@ scan_devices(fw_handle_t handle)
 		if (de == NULL)
 			break;
 
-		if (strncmp(de->d_name,
-			    FW_DEVICE_PREFIX, strlen(FW_DEVICE_PREFIX)) != 0)
+		if (!is_fw_device_name(de->d_name))
 			continue;
 
 		snprintf(filename, sizeof filename, FW_DEVICE_DIR "/%s", de->d_name);
@@ -259,6 +256,7 @@ handle_device_event(raw1394handle_t handle,
 	struct device *device = (struct device *) ec;
 	struct address_closure *ac;
 	struct request_closure *rc;
+	unsigned long tag;
 	raw1394_errcode_t errcode;
 	int len, phy_id;
 	int i;
@@ -297,13 +295,22 @@ handle_device_event(raw1394handle_t handle,
 			memcpy(rc->data, u->response.data, rc->length);
 
 		errcode = fw_to_raw1394_errcode(u->response.rcode);
+		tag = rc->tag;
+		free(rc);
 
-		return fwhandle->tag_handler(handle, rc->tag, errcode);
+		return fwhandle->tag_handler(handle, tag, errcode);
 
 	case FW_CDEV_EVENT_REQUEST:
 		ac = u64_to_ptr(u->request.closure);
 		return ac->callback(handle, ac, &u->request, i);
 
+#ifdef FW_CDEV_EVENT_ISO_RESOURCE_ALLOCATED /* added in kernel 2.6.30 */
+	case FW_CDEV_EVENT_ISO_RESOURCE_ALLOCATED:
+	case FW_CDEV_EVENT_ISO_RESOURCE_DEALLOCATED:
+		memcpy(u64_to_ptr(u->iso_resource.closure), u,
+		       sizeof u->iso_resource);
+		return 0;
+#endif
 	default:
 	case FW_CDEV_EVENT_ISO_INTERRUPT:
 		/* Never happens. */
@@ -327,8 +334,7 @@ handle_inotify(raw1394handle_t handle, struct epoll_closure *ec,
 	len = read(fwhandle->inotify_fd, event, BUFFER_SIZE);
 	if (!(event->mask & IN_CREATE))
 		return -1;
-	if (strncmp(event->name,
-		    FW_DEVICE_PREFIX, strlen(FW_DEVICE_PREFIX)) != 0)
+	if (!is_fw_device_name(event->name))
 		return 0;
 	snprintf(filename, sizeof filename, FW_DEVICE_DIR "/%s", event->name);
 	fd = open(filename, O_RDWR);
@@ -641,6 +647,7 @@ int fw_set_port(fw_handle_t handle, int port)
 			handle->local_device = &handle->devices[i];
 
 		handle->generation = reset.generation;
+		handle->abi_version = get_info.version;
 
 		i++;
 	}
@@ -945,7 +952,29 @@ send_request(fw_handle_t handle, int tcode,
 {
 	struct fw_cdev_send_request *request;
 	struct request_closure *closure;
-	int i;
+	int ioctl_nr = FW_CDEV_IOC_SEND_REQUEST;
+	int fd, i, retval;
+
+#ifdef FW_CDEV_IOC_SEND_STREAM_PACKET /* added in kernel 2.6.30 */
+	if (tcode == TCODE_STREAM_DATA) {
+		ioctl_nr = FW_CDEV_IOC_SEND_STREAM_PACKET;
+		fd = handle->ioctl_fd;
+	}
+#else
+	if (tcode == TCODE_STREAM_DATA) {
+		errno = ENOSYS;
+		return -1;
+	}
+#endif
+
+#ifdef FW_CDEV_IOC_SEND_BROADCAST_REQUEST /* added in kernel 2.6.30 */
+	if (node == 0xffff) {
+		ioctl_nr = FW_CDEV_IOC_SEND_BROADCAST_REQUEST;
+		fd = handle->ioctl_fd;
+	}
+#endif
+	if (ioctl_nr != FW_CDEV_IOC_SEND_REQUEST)
+		goto node_id_ok;
 
 	if (node > handle->reset.root_node_id) {
 		handle->err = -RCODE_NO_ACK;
@@ -965,6 +994,9 @@ send_request(fw_handle_t handle, int tcode,
 		errno = fw_errcode_to_errno(handle->err);
 		return -1;
 	}
+	fd = handle->devices[i].fd;
+
+node_id_ok:
 
 	closure = malloc(sizeof *closure);
 	if (closure == NULL) {
@@ -973,19 +1005,39 @@ send_request(fw_handle_t handle, int tcode,
 		return -1;
 	}
 
-	closure->data = out;
+	closure->data   = out;
 	closure->length = out_length;
-	closure->tag = tag;
+	closure->tag    = tag;
 
 	request = (struct fw_cdev_send_request *) handle->buffer;
-	request->tcode = tcode;
+	request->tcode      = tcode;
+	request->length     = in_length > out_length ? in_length : out_length;
+	request->offset     = addr;
+	request->closure    = ptr_to_u64(closure);
+	request->data       = ptr_to_u64(in);
 	request->generation = handle->generation;
-	request->offset = addr;
-	request->length = in_length > out_length ? in_length : out_length;
-	request->closure = ptr_to_u64(closure);
-	request->data = ptr_to_u64(in);
 
-	return ioctl(handle->devices[i].fd, FW_CDEV_IOC_SEND_REQUEST, request);
+#ifdef FW_CDEV_IOC_SEND_STREAM_PACKET
+	if (tcode == TCODE_STREAM_DATA) {
+		struct fw_cdev_send_stream_packet *p
+		    = (struct fw_cdev_send_stream_packet *) request;
+
+		p->length     = in_length;
+		p->tag        = (addr >> 14) & 0x3;
+		p->channel    = (addr >> 8) & 0x3f;
+		p->sy         = addr & 0xf;
+		p->closure    = ptr_to_u64(closure);
+		p->data       = ptr_to_u64(in);
+		p->generation = handle->generation;
+		p->speed      = (addr >> 4) & 0x7;
+	}
+#endif
+
+	retval = ioctl(fd, ioctl_nr, request);
+	if (retval < 0)
+		free(closure);
+
+	return retval;
 }
 
 int
@@ -1104,9 +1156,13 @@ fw_start_async_stream(fw_handle_t handle, unsigned int channel,
 			   unsigned int speed, size_t length, quadlet_t *data,
 			   unsigned long rawtag)
 {
-	/* FIXME: implement this? */
-	errno = ENOSYS;
-	return -1;
+	nodeaddr_t addr = (tag     & 0x3)  << 14 |
+			  (channel & 0x3f) <<  8 |
+			  (speed   & 0x7)  <<  4 |
+			  (sy      & 0xf);
+
+	return send_request(handle, TCODE_STREAM_DATA,
+			    0, addr, length, data, 0, NULL, rawtag);
 }
 
 
@@ -1231,13 +1287,17 @@ fw_lock64(raw1394handle_t handle, nodeid_t node, nodeaddr_t addr,
 }
 
 int
-fw_async_stream(fw_handle_t handle, unsigned int channel,
+fw_async_stream(raw1394handle_t handle, unsigned int channel,
 		     unsigned int tag, unsigned int sy, unsigned int speed,
 		     size_t length, quadlet_t *data)
 {
-	/* FIXME: implement this? */
-	errno = ENOSYS;
-	return -1;
+	nodeaddr_t addr = (tag     & 0x3)  << 14 |
+			  (channel & 0x3f) <<  8 |
+			  (speed   & 0x7)  <<  4 |
+			  (sy      & 0xf);
+
+	return send_request_sync(handle, TCODE_STREAM_DATA,
+				 0, addr, length, data, 0, NULL);
 }
 
 int
@@ -1321,3 +1381,90 @@ fw_get_config_rom(fw_handle_t handle, quadlet_t *buffer,
 
 	return 0;
 }
+
+#ifdef FW_CDEV_IOC_ALLOCATE_ISO_RESOURCE_ONCE /* added in kernel 2.6.30 */
+
+static int
+iso_resource_modify(raw1394handle_t handle, unsigned int bandwidth,
+		    int channel, enum raw1394_modify_mode mode)
+{
+	fw_handle_t fwhandle = handle->mode.fw;
+	struct fw_cdev_allocate_iso_resource resource;
+	struct fw_cdev_event_iso_resource event;
+	int ioctl_nr;
+	int err;
+
+	if (channel > 63) {
+		errno = EINVAL;
+		return -1;
+	}
+
+	event.closure = 0;
+	event.channel = -1;
+	event.bandwidth = 0;
+
+	resource.closure = ptr_to_u64(&event);
+	resource.channels = channel >= 0 ? 1ULL << channel : 0;
+	resource.bandwidth = bandwidth;
+
+	ioctl_nr = mode == RAW1394_MODIFY_ALLOC ?
+		   FW_CDEV_IOC_ALLOCATE_ISO_RESOURCE_ONCE :
+		   FW_CDEV_IOC_DEALLOCATE_ISO_RESOURCE_ONCE;
+
+	err = ioctl(fwhandle->ioctl_fd, ioctl_nr, &resource);
+
+	while (err >= 0 && event.closure != resource.closure)
+		err = fw_loop_iterate(handle);
+
+	if (err < 0)
+		return err;
+
+	if ((channel >= 0 && event.channel < 0) ||
+	    (bandwidth > 0 && event.bandwidth == 0)) {
+		errno = EIO;
+		return -1;
+	}
+
+	return 0;
+}
+
+static inline int abi_v2_available(raw1394handle_t handle)
+{
+	return handle->mode.fw->abi_version >= 2;
+}
+
+#else
+
+static inline int
+iso_resource_modify(raw1394handle_t handle, unsigned int bandwidth,
+		    int channel, enum raw1394_modify_mode mode)
+{
+	return -1;
+}
+
+static inline int abi_v2_available(raw1394handle_t handle)
+{
+	return 0;
+}
+
+#endif  /* defined(FW_CDEV_IOC_ALLOCATE_ISO_RESOURCE_ONCE) */
+
+int
+fw_bandwidth_modify(raw1394handle_t handle, unsigned int bandwidth,
+		    enum raw1394_modify_mode mode)
+{
+	if (abi_v2_available(handle))
+		return iso_resource_modify(handle, bandwidth, -1, mode);
+	else
+		return ieee1394_bandwidth_modify(handle, bandwidth, mode);
+}
+
+int
+fw_channel_modify(raw1394handle_t handle, unsigned int channel,
+		  enum raw1394_modify_mode mode)
+{
+	if (abi_v2_available(handle))
+		return iso_resource_modify(handle, 0, channel, mode);
+	else
+		return ieee1394_channel_modify(handle, channel, mode);
+}
